import os
import gc

import json
import pandas as pd
import gzip

from gensim.models import TfidfModel, LsiModel
from gensim.corpora import Dictionary

from nlpia.data import BIGDATA_PATH, read_csv

KEEP_N = 300000  # max vocab size
NO_BELOW = 5      # min DF (count)
NO_ABOVE = .7     # max DF (fraction)

# Only 5 of these tokens are saved for a no_below=2 filter: PyCons NLPS #PyCon2016 #NaturalLanguageProcessing #naturallanguageprocessing
cased_tokens = 'PyConOpenSpaces PyCon PyCon2017 PyCon2018 PyCon2016 PyCon2015 OpenSpace PyconTutorial'.split()
cased_tokens += 'NLP NaturalLanguageProcessing NLPInAction NaturalLanguageProcessingInAction NLPIA Twote Twip'.split()
cased_tokens += [s + 's' for s in cased_tokens]
cased_tokens += 'TotalGood TotalGoods HobsonLane Hob Hobs TotalGood.com www.TotalGood.com http://www.TotalGood.com https://www.TotalGood.com'.split()
allcase_tokens = cased_tokens + [s.lower() for s in cased_tokens]
allcase_tokens += [s.title() for s in cased_tokens]
allcase_tokens += [s.upper() for s in cased_tokens]
KEEP_TOKENS = allcase_tokens + ['#' + s for s in allcase_tokens]

# takes 15 minutes and 10GB of RAM for 500k tweets if you keep all 20M unique tokens/names URLs
vocab_path = os.path.join(BIGDATA_PATH, 'vocab939370.pkl')
if os.path.isfile(vocab_path):
    print('Loading vocab: {} ...'.format(vocab_path))
    vocab = Dictionary.load(vocab_path)
    print(' len(vocab) loaded: {}'.format(len(vocab.dfs)))
else:
    tweets_path = os.path.join(BIGDATA_PATH, 'tweets.csv.gz')
    print('Loading tweets: {} ...'.format(tweets_path))
    tweets = read_csv(tweets_path)
    tweets = pd.np.array(tweets.text.str.split())
    with gzip.open(os.path.join(BIGDATA_PATH, 'tweets.txt.gz'), 'w') as f:
        for tokens in tweets:
            f.write((' '.join(tokens) + '\n').encode('utf-8'))
    # tweets['text'] = tweets.text.apply(lambda s: eval(s).decode('utf-8'))
    # tweets['user'] = tweets.user.apply(lambda s: eval(s).decode('utf-8'))
    # tweets.to_csv('tweets.csv.gz', compression='gzip')
    print('Computing vocab from {} tweets...'.format(len(tweets)))
    vocab = Dictionary(tweets, no_below=NO_BELOW, no_above=NO_ABOVE, keep_tokens=set(KEEP_TOKENS))

vocab.filter_extremes(no_below=NO_BELOW, no_above=NO_ABOVE, keep_n=KEEP_N, keep_tokens=set(KEEP_TOKENS))
print(' len(vocab) after filtering: {}'.format(len(vocab.dfs)))


# no time at all, just a bookeeping step, doesn't actually compute anything
tfidf = TfidfModel(id2word=vocab, dictionary=vocab)
tfidf.save(os.path.join(BIGDATA_PATH, 'tfidf{}.pkl'.format(len(vocab.dfs))))

tweets = [vocab.doc2bow(tw) for tw in tweets]
json.dump(tweets, gzip.open(os.path.join(BIGDATA_PATH, 'tweet_bows.json.gz'), 'w'))

gc.collect()

# LSA is more useful name than LSA
lsa = LsiModel(tfidf[tweets], num_topics=200, id2word=vocab, extra_samples=100, power_iters=2)

# these models can be big
lsa.save(os.path.join(BIGDATA_PATH, 'lsa_tweets'))



#!usr/bin/env python3
from __future__ import print_function, unicode_literals, division, absolute_import
from builtins import int, round, str,  object  # noqa
from future import standard_library
standard_library.install_aliases()  # noqa: Counter, OrderedDict, 
from past.builtins import basestring   # noqa:

import pandas as pd

import future        # noqa
import builtins      # noqa
import past          # noqa
import six           # noqa
from lshash.lshash import LSHash

np = pd.np


def sphere():
    X = np.random.normal(size=(1000, 3))
    lsh = LSHash(10, 3, num_hashtables=5)
    for x in X:
        x /= np.linalg.norm(x)
        lsh.index(x)
    closest = lsh.query(X[0] + np.array([-0.001, 0.001, -0.001]), distance_func="cosine")
    assert len(closest) >= 10
    assert 0.05 >= closest[9][-1] > 0.0003


def hyperspheres(X=16, num_samples=200000):
    """ Demonstrate curse of dimensionality and where LSH starts to fail

    Returns:
      lsh, X, secondclosest, tenthclosest

    >>> import pandas as pd
    >>> lsh, vectors, dfs = hyperspheres(16)
    >>> for df in dfs:
    ...     print(df)
    """
    X = np.random.uniform(size=(num_samples, X)) if isinstance(X, int) else X
    closest = []
    secondclosest = []
    tenthclosest = []
    hundredthclosest = []
    for D in range(2, X.shape[1] + 1):
        lsh = LSHash(int(64 / D) + D, D, num_hashtables=D)

        # query vector
        q = np.random.uniform(size=(D,))
        q /= np.linalg.norm(q)

        distances = []
        for x in X[:, :D]:
            x /= np.linalg.norm(x)
            distances += [1. - np.sum(x * q)]  # cosine similarity
            lsh.index(x)
        distances = sorted(distances)
        print(distances[:10])
        closest10 = lsh.query(q, distance_func='cosine')

        N = len(closest10)
        hundredthclosest += [[D, N, closest10[min(99, N - 1)][-1] if N else 2., distances[min(99, N - 1)]]]
        tenthclosest += [[D, N, closest10[min(9, N - 1)][-1] if N else 2., distances[min(9, N - 1)]]]
        secondclosest += [[D, N, closest10[min(1, N - 1)][-1] if N else 2., distances[min(1, N - 1)]]]
        closest += [[D, N, closest10[0][-1] if N else 2., distances[0]]]
        print("is correct: 100th 10th 2nd 1st")
        print(round(hundredthclosest[-1][-1], 14) == round(hundredthclosest[-1][-2], 14))
        print(round(tenthclosest[-1][-1], 14) == round(tenthclosest[-1][-2], 14))
        print(round(secondclosest[-1][-1], 14) == round(secondclosest[-1][-2], 14))
        print(round(closest[-1][-1], 14) == round(closest[-1][-2], 14))
        print("distances: 100th 10th 2nd 1st")
        print(hundredthclosest[-1])
        print(tenthclosest[-1])
        print(secondclosest[-1])
        print(closest[-1])
    dfs = []
    for k, (i, df) in enumerate(zip([100, 10, 2, 1], [hundredthclosest, tenthclosest, secondclosest, closest])):
        df = pd.DataFrame(df, columns='D N dist{} true_dist{}'.format(i, i).split()).round(14)
        df['correct{}'.format(i)] = df['dist{}'.format(i)] == df['true_dist{}'.format(i)]
        dfs += [df]
    # for i, tc in enumerate(tenthclosest):
    #     assert 1e-9 < tc[-2] or 1e-6 < 0.2
    return lsh, X, dfs



from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases() # noqa
from builtins import *  # noqa

# from gensim.models import Word2Vec
from gensim import corpora
from gensim import utils

from nlpia.constants import logging
logger = logging.getLogger(__name__)


def tokens2ngrams(tokens, n=2):
    tokens = list(tokens)
    ngrams = []
    for i in range(len(tokens) - n + 1):
        ngrams.append(' '.join(tokens[i:i + n]))
    return ngrams


def passthrough(*args, **kwargs):
    return args[0] if len(args) else list(kwargs.values())[0]


def return_false(*args, **kwargs):
    return False


def return_true(*args, **kwargs):
    return True


def noop(*args, **kwargs):
    pass


def return_none(*args, **kwargs):
    pass


def to_unicode(sorb, allow_eval=False):
    """Ensure that strings are unicode (UTF-8 encoded).

    Evaluate bytes literals that are sometimes accidentally created by str(b'whatever')

    >>> to_unicode(b'whatever')
    'whatever'
    >>> to_unicode(b'b"whatever"')
    "b'whatever'"
    >>> '"{}"'.format(b'whatever')
    '"b\'whatever\'"'
    >>> str(b'wat')
    "b'wat'"
    >>> to_unicode(str(b'whatever'))
    'whatever'
    """
    if isinstance(sorb, bytes):
        sorb = sorb.decode('utf-8')
    if sorb and (sorb[:2] == "b'" and sorb[-1] == "'") or (sorb[:2] == 'b"' and sorb[-1] == '"'):
        sorb = eval(sorb, {'__builtins__': None}, {})
    return str(sorb)


class TweetCorpus(corpora.TextCorpus):
    ignore_matcher = return_none   # compiled regular expression for token matches to skip/ignore
    num_grams = 2
    case_normalizer = str
    tokenizer = None
    mask = None

    def get_texts(self):
        """ Parse documents from a .txt file assuming 1 document per line, yielding lists of filtered tokens """
        with self.getstream() as text_stream:
            for i, line in enumerate(text_stream):
                line = to_unicode(line)
                line = (TweetCorpus.case_normalizer or passthrough)(line)
                # line = self.case_normalizer(line)
                if self.mask is not None and not self.mask[i]:
                    continue
                ngrams = []
                for ng in tokens2ngrams((TweetCorpus.tokenizer or str.split)(line), n=self.num_grams):
                    if self.ignore_matcher(ng):
                        continue
                    ngrams += [ng]
                if not (i % 1000):
                    print(line)
                    print(ngrams)
                yield ngrams

    def __len__(self):
        """ Enables `len(corpus)` """
        if 'length' not in self.__dict__:
            logger.info("Computing the number of lines in the corpus size (calculating number of documents)")
            self.length = sum(1 for doc in self.getstream())
        return self.length


class SMSCorpus(corpora.TextCorpus):
    ignore_matcher = return_none   # compiled regular expression for token matches to skip/ignore
    num_grams = 2
    case_normalizer = utils.to_unicode
    tokenizer = str.split
    mask = None

    def get_texts(self):
        """ Parse documents from a .txt file assuming 1 document per line, yielding lists of filtered tokens """
        with self.getstream() as text_stream:
            for i, line in enumerate(text_stream):
                line = SMSCorpus.case_normalizer(line)
                if self.mask is not None and not self.mask[i]:
                    continue
                ngrams = []
                for ng in tokens2ngrams(self.tokenizer(line)):
                    if SMSCorpus.ignore_matcher(ng):
                        continue
                    ngrams += [ng]
                yield ngrams

    def __len__(self):
        """ Enables `len(corpus)` """
        if 'length' not in self.__dict__:
            logger.info("Computing the number of lines in the corpus size (calculating number of documents)")
            self.length = sum(1 for doc in self.getstream())
        return self.length



from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa
from builtins import *  # noqa

from future.utils import viewitems  # noqa
from past.builtins import basestring  # noqa

import requests

from pugnlp import regexes

# amazon product reviews for recommendation engine training and review sentiment analysis
response = requests.get('http://jmcauley.ucsd.edu/data/amazon/')
urls_product_review = [m[0] for m in regexes.cre_url_popular.findall(response.text) if m[0].lower().endswith('.json.gz')]

response = requests.get('http://jmcauley.ucsd.edu/data/amazon/qa/')
urls_question_answer = [m[1] for m in regexes.cre_href.findall(response.text) if m[1].lower().endswith('.json.gz')]


with open('download_mccauley_autogenerated.sh', 'w') as f:
    for pr in urls_product_review:
        f.write('wget ' + pr)
    for qa in urls_question_answer:
        f.write('wget ' + qa)



from __future__ import absolute_import, division, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()  # noqa
from builtins import *  # noqa
import tempfile

import pandas as pd
from nlpia.constants import logging

from annoy import AnnoyIndex

np = pd.np
logger = logging.getLogger(__name__)


def representative_sample(X, num_samples, save=False):
    """Sample vectors in X, prefering edge cases and vectors farthest from other vectors in sample set


    """
    X = X.values if hasattr(X, 'values') else np.array(X)
    N, M = X.shape
    rownums = np.arange(N)
    np.random.shuffle(rownums)

    idx = AnnoyIndex(M)
    for i, row in enumerate(X):
        idx.add_item(i, row)
    idx.build(int(np.log2(N)) + 1)

    if save:
        if isinstance(save, basestring):
            idxfilename = save
        else:
            idxfile = tempfile.NamedTemporaryFile(delete=False)
            idxfile.close()
            idxfilename = idxfile.name
        idx.save(idxfilename)
        idx = AnnoyIndex(M)
        idx.load(idxfile.name)

    samples = -1 * np.ones(shape=(num_samples,), dtype=int)
    samples[0] = rownums[0]
    # FIXME: some integer determined by N and num_samples and distribution
    j, num_nns = 0, min(1000, int(num_samples / 2. + 1))
    for i in rownums:
        if i in samples:
            continue
        nns = idx.get_nns_by_item(i, num_nns)
        # FIXME: pick vector furthest from past K (K > 1) points or outside of a hypercube (sized to uniformly fill the space) around the last sample
        try:
            samples[j + 1] = np.setdiff1d(nns, samples)[-1]
        except:
            samples[j + 1]
        if len(num_nns) < num_samples / 3.:
            num_nns = min(N, 1.3 * num_nns)
        j += 1



from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases() # noqa
from builtins import *  # noqa

import os
import gc

import json
import pandas as pd
import gzip
from tqdm import tqdm

from gensim.models import TfidfModel, LsiModel
from gensim.corpora import Dictionary

from nlpia.data import BIGDATA_PATH, read_csv
from nlpia.gensim_utils import TweetCorpus

KEEP_N = 300000   # max vocab size
NO_BELOW = 5      # min DF (count)
NO_ABOVE = .99    # max DF (fraction)

# Only 5 of these tokens are saved for a no_below=2 filter: PyCons NLPS #PyCon2016 #NaturalLanguageProcessing #naturallanguageprocessing
cased_tokens = 'PyConOpenSpaces PyCon PyCon2017 PyCon2018 PyCon2016 PyCon2015 OpenSpace PyconTutorial'.split()
cased_tokens += 'NLP NaturalLanguageProcessing NLPInAction NaturalLanguageProcessingInAction NLPIA Twote Twip'.split()
cased_tokens += [s + 's' for s in cased_tokens]
allcase_tokens = cased_tokens + [s.lower() for s in cased_tokens]
allcase_tokens += [s.title() for s in cased_tokens]
allcase_tokens += [s.upper() for s in cased_tokens]
KEEP_TOKENS = allcase_tokens + ['#' + s for s in allcase_tokens]

# takes 15 minutes and 10GB of RAM for 500k tweets if you keep all 20M unique tokens/names URLs
vocab_path = os.path.join(BIGDATA_PATH, 'vocab.gensim')
if os.path.isfile(vocab_path):
    print('Loading vocab: {} ...'.format(vocab_path))
    vocab = Dictionary.load(vocab_path)
    print(' len(vocab) loaded: {}'.format(len(vocab.dfs)))
else:
    tokens_path = os.path.join(BIGDATA_PATH, 'tweets.txt.gz')
    if not os.path.isfile(tokens_path):
        tweets_path = os.path.join(BIGDATA_PATH, 'tweets.csv.gz')
        print('Loading tweets: {} ...'.format(tweets_path))
        tweets = read_csv(tweets_path)
        tweets = pd.np.array(tweets.text.apply(eval).apply(bytes.decode).str.split())
        with gzip.open(tokens_path, 'wb') as f:
            for tokens in tqdm(tweets):
                f.write((b' '.join(tokens) + b'\n'))
    # tweets['text'] = tweets.text.apply(lambda s: eval(s).decode('utf-8'))
    # tweets['user'] = tweets.user.apply(lambda s: eval(s).decode('utf-8'))
    # tweets.to_csv('tweets.csv.gz', compression='gzip')
    print('Computing vocab from tweets in {} ...'.format(tokens_path))
    corpus = TweetCorpus()
    corpus.input = gzip.open(tokens_path, 'rb')
    vocab = Dictionary(corpus, prune_at=20000000)
    vocab.filter_extremes(no_below=NO_BELOW, no_above=NO_ABOVE, keep_tokens=set(KEEP_TOKENS))

vocab.filter_extremes(no_below=NO_BELOW, no_above=NO_ABOVE, keep_n=KEEP_N, keep_tokens=set(KEEP_TOKENS))
print(' len(vocab) after filtering: {}'.format(len(vocab.dfs)))


# no time at all, just a bookeeping step, doesn't actually compute anything
tfidf = TfidfModel(id2word=vocab, dictionary=vocab)
tfidf.save(os.path.join(BIGDATA_PATH, 'tfidf{}.pkl'.format(len(vocab.dfs))))

tweets = [vocab.doc2bow(tw) for tw in tweets]
json.dump(tweets, gzip.open(os.path.join(BIGDATA_PATH, 'tweet_bows.json.gz'), 'w'))

gc.collect()

# LSA is more useful name than LSA
lsa = LsiModel(tfidf[tweets], num_topics=200, id2word=vocab, one_pass=False, extra_samples=150, power_iters=3)

# these models can be big
lsa.save(os.path.join(BIGDATA_PATH, 'lsa_tweets'))



#!/usr/bin/env python3
"""model_poly_tsne

Run nlpia.data.download() to download GBs of models like W2V and the LSAmodel used here

Computes a TSNE embedding for the tweet LSA model and then fit a 2nd degree polynomial to that embedding.
"""

import os
import gc

import pandas as pd
from tqdm import tqdm

from gensim.models import LsiModel, TfidfModel

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
# from sklearn.svm import SVR
from sklearn.manifold import TSNE
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import cross_val_score

from nlpia.constants import BIGDATA_PATH
from nlpia.data import read_csv

import sklearn.metrics.pairwise

np = pd.np


def positive_projection(x, y, max_norm=1.0):
    proj = max_norm - float(np.dot(x, y))
    if proj < 1e-15:
        print(x, y, proj)
    return max(proj, 0.0) ** 0.5


def positive_distances(X, metric='cosine'):
    X = X.values if (hasattr(X, 'values') and not callable(X.values)) else X
    metric = getattr(sklearn.metrics.pairwise, metric + '_distances') if isinstance(metric, basestring) else metric
    distances = metric(X)
    distances[distances < 0] = 0.0
    return distances


def bent_distances(X, y, weight=1.0, metric='cosine'):
    y = np.array(y).reshape((len(X), 1))
    distances = positive_distances(X, metric=metric)
    distances += weight * sklearn.metrics.pairwise.euclidean_distances(np.matrix(y).T)
    return distances


def train_tsne(training_size=2000, metric='cosine', n_components=3, perplexity=100, angle=.12):
    # adjust this downward to see it it affects accuracy
    np = pd.np

    tweets = read_csv(os.path.join(BIGDATA_PATH, 'tweets.csv.gz'))
    tweets = tweets[tweets.isbot >= 0]
    gc.collect()  # reclaim RAM released above

    # labels3 = tweets.isbot.apply(lambda x: int(x * 3))
    labels = tweets.isbot.apply(lambda x: int(x * 2))

    lsa = LsiModel.load(os.path.join(BIGDATA_PATH, 'lsa_tweets_5589798_2003588x200.pkl'))
    tfidf = TfidfModel(id2word=lsa.id2word, dictionary=lsa.id2word)
    bows = np.array([lsa.id2word.doc2bow(txt.split()) for txt in tweets.text])
    # tfidfs = tfidf[bows]

    X = pd.DataFrame([pd.Series(dict(v)) for v in tqdm(lsa[tfidf[bows]], total=len(bows))], index=tweets.index)

    mask = ~X.isnull().any(axis=1)
    mask.index = tweets.index
    # >>> sum(~mask)
    # 99
    # >>> tweets.loc[mask.argmin()]
    # isbot                 0.17
    # strict                  13
    # user      b'CrisParanoid:'
    # text         b'#sad again'
    # Name: 571, dtype: object

    X = X[mask]
    y = tweets.isbot[mask]
    labels = labels[mask]

    test_size = 1.0 - training_size if training_size < 1 else float(len(X) - training_size) / len(X)
    Xindex, Xindex_test, yindex, yindex_test = train_test_split(X.index.values, y.index.values, test_size=test_size)
    X, Xtest, y, ytest = X.loc[Xindex], X.loc[Xindex_test], y.loc[yindex], y.loc[yindex_test]
    labels_test = labels.loc[yindex_test]
    labels = labels.loc[yindex]

    tsne = TSNE(metric='precomputed', n_components=n_components, angle=angle, perplexity=perplexity)
    tsne = tsne.fit(positive_distances(X.values, metric=metric))

    return tsne, X, Xtest, y, ytest


def embedding_correlation(Xtest, ytest):
    pass


def plot_embedding(tsne, labels, index=None):
    labels = labels.values if (hasattr(labels, 'values') and not callable(labels.values)) else labels
    colors = np.array(list('gr'))[labels]
    df = pd.DataFrame(tsne.embedding_, columns=list('xy'), index=index)
    return df.plot(kind='scatter', x='x', y='y', c=colors)



from seaborn import plt
from mpl_toolkits.mplot3d import   # noqa

import pandas as pd

from plyfile import PlyData

np = pd.np


horse = PlyData.read('horse.ply')
points = np.array(horse.elements[0].data)
# facets = np.array(horse.elements[1].data)
points = np.array([row.tolist() for row in points])
# pd.DataFrame(points, columns='x y z alpha r g b'.split())
df = pd.DataFrame(points[:, :3], columns='x y z'.split())
# df.to_csv('horse.csv')


h = pd.read_csv('horse.csv', header=None).values[:, :3]
h = pd.DataFrame(h, columns='x y z'.split())
h = h.sample(1000)

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.scatter(h.x, h.y, h.z, zdir='z', s=20, c=None, depthshade=True)
plt.show()



""" Explore the "topology" of High-D space and make anolgies in 2-3D

>>> from nlpia.plots import *
>>> df = pd.DataFrame(np.random.uniform(size=(1000,3)))
>>> labels = df.ix[:,:3].dot(np.array([3,-4,0])/5.) > 0
>>> df['c'] = labels.astype(int)
>>> scatter_3d(df, labels='c')
>>> plt.show()
"""
from seaborn import plt

import pandas as pd

from nlpia.plots import scatter_3d

np = pd.np


df = pd.DataFrame(pd.np.random.randn(1000, 10))

np = pd.np
plt.figure()
f = plt.figure()
ax = f.add_subplot(111)
axes = df.plot(kind='scatter', x=0, y=1, ax=ax)
plt.title('2D Normally Distributed')
plt.tight_layout()
plt.show()

scatter_3d(df)
plt.show()
df.values.dot(df.values)
np.do(df.values, df.values)
np.dot(df.values, df.values)
np.dot(df.values, df.values.T)
np.dot(df.values, df.values.T).shape
np.dot(df.values, df.values.T).shape.sum()
np.dot(df.values.T, df.values)
np.dot(df.values.T, df.values).shape
norms = pd.DataFrame([[np.linalg.norm(row[:i]) for i in range(2,11)] for row in df])
row
for row in df:
    print(row)
    break
df
norms = pd.DataFrame([[np.linalg.norm(row[:i]) for i in range(2,11)] for j, row in df.iterrows()])
norms
plt.hist(norms.values[:,9])
plt.hist(norms.values[:,8])
plt.show()
stds = []
for i in range(9):
    plt.hist(norms.values[:,i])
    plt.tight_layout()
    plt.show()
stds = []
for i in range(9):
    plt.hist(norms.values[:,i])
    plt.tight_layout()
pd.DataFrame(norms).describe()
hist
df
df = pd.DataFrame(np.random.uniform(1000,30))
df = pd.DataFrame(np.random.uniform(shape=(1000,30)))
df = pd.DataFrame(np.random.uniform((1000,30)))
df
df = pd.DataFrame(np.random.uniform(1000))
np.random.uniform(1000)
np.random.uniform?
df = pd.DataFrame(np.random.uniform(size=(1000,30)))


scatter_3d(df)
plt.show()
norms = pd.DataFrame([[np.linalg.norm(row[:i]) for i in range(2,21)] for j, row in df.iterrows()])
stds = []
for i in range(19):
    plt.hist(norms.values[:,i])
    plt.tight_layout()
plt.show()
axes = pd.DataFrame(pd.np.random.uniform(size=(1000, 2)), columns=list('xy')).plot(kind='scatter', x='x', y='y', ax=ax)
axes.tight_layout()
plt.show()
axes.plot()
axes = pd.DataFrame(pd.np.random.uniform(size=(1000, 2)), columns=list('xy')).plot(kind='scatter', x='x', y='y', ax=ax)
plt.show()
f = plt.figure()
ax = f.add_subplot(111)
axes = pd.DataFrame(pd.np.random.uniform(size=(1000, 2)), columns=list('xy')).plot(kind='scatter', x='x', y='y', ax=ax)
plt.tight_layout()
plt.show()
df = pd.DataFrame(np.random.uniform(size=(1000,60)))
norms = pd.DataFrame([[np.linalg.norm(row[:i]) for i in range(2,4,61)] for j, row in df.iterrows()])
stds = []
for i in range(15):
    plt.hist(norms.values[:,i])
    plt.tight_layout()
plt.show()
norms = pd.DataFrame([[np.linalg.norm(row[:i]) for i in range(2,4,61)] for j, row in df.iterrows()])
norms.describe()
range(2,61,4)
list(range(2,61,4))
norms = pd.DataFrame([[np.linalg.norm(row[:i]) for i in range(2,61,4)] for j, row in df.iterrows()])
norms.describe()
stds = []
for i in range(15):
    plt.hist(norms.values[:,i])
    plt.tight_layout()
plt.show()
!git pull
hist




from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa: Counter, OrderedDict, 
from builtins import *  # noqa
from past.builtins import basestring   # noqa

from seaborn import plt
from mpl_toolkits.mplot3d import Axes3D  # noqa

import pandas as pd
np = pd.np


h = pd.read_csv('pointcloud.csv.gz', header=None).values[:, :3]
h = pd.DataFrame(h, columns='x y z'.split())
h = h.sample(1000)

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.scatter(h.x, h.y, h.z, zdir='z', s=20, c=None, depthshade=True)
plt.show()



#!/usr/bin/env python3
"""model_isbot

Run nlpia.data.download() to download GBs of models like W2V and the LSAmodel used here
"""
from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases() # noqa
from builtins import *  # noqa

import os

import pandas as pd
from tqdm import tqdm
import gc

from gensim.models import LsiModel, TfidfModel

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.manifold import TSNE
# from sklearn.svm import SVR
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import cross_val_score

from nlpia.constants import BIGDATA_PATH
from nlpia.data import read_csv

from nlpia.embedders import positive_distances

# adjust this downward
np = pd.np


def train_lda(training_size=2000, metric='cosine'):
    tweets = read_csv(os.path.join(BIGDATA_PATH, 'tweets.csv.gz'))
    tweets = tweets[tweets.isbot >= 0]

    # labels3 = tweets.isbot.apply(lambda x: int(x * 3))
    labels = tweets.isbot.apply(lambda x: int(x * 2))

    lsa = LsiModel.load(os.path.join(BIGDATA_PATH, 'lsa_tweets_5589798_2003588x200.pkl'))
    tfidf = TfidfModel(id2word=lsa.id2word, dictionary=lsa.id2word)
    bows = np.array([lsa.id2word.doc2bow(txt.split()) for txt in tweets.text])
    # tfidfs = tfidf[bows]

    X = pd.DataFrame([pd.Series(dict(v)) for v in tqdm(lsa[tfidf[bows]], total=len(bows))], index=tweets.index)
    mask = ~X.isnull().any(axis=1)
    mask.index = tweets.index
    X = X[mask]
    y = tweets.isbot[mask]
    labels = labels[mask]
    # labels3 = labels3[mask]

    test_size = 1.0 - training_size if training_size < 1 else float(len(X) - training_size) / len(X)
    Xindex, Xindex_test, yindex, yindex_test = train_test_split(X.index.values, y.index.values, test_size=test_size)
    X, Xtest, y, ytest = X.loc[Xindex], X.loc[Xindex_test], y.loc[yindex], y.loc[yindex_test]
    labels_test = labels.loc[yindex_test]
    labels = labels.loc[yindex]

    lda = LDA('lsqr', 'auto', n_components=3)
    print(cross_val_score(lda, Xtest, labels_test, cv=7))

    lda = LDA('lsqr', 'auto', n_components=3)
    lda = lda.fit(X.values, labels.values)
    y_lda = lda.predict(Xtest)
    print(mean_squared_error(y_lda, ytest))

    df_test = pd.DataFrame(lda.predict(Xtest), index=Xtest.index, columns=['predict'])
    df_test['truth'] = labels_test
    return lda, df_test


def train_tsne(training_size=2000, metric='cosine', n_components=3, perplexity=100, angle=.12):
    # adjust this downward to see it it affects accuracy
    np = pd.np

    tweets = read_csv(os.path.join(BIGDATA_PATH, 'tweets.csv.gz'))
    tweets = tweets[tweets.isbot >= 0]
    gc.collect()  # reclaim RAM released above

    # labels3 = tweets.isbot.apply(lambda x: int(x * 3))
    labels = tweets.isbot.apply(lambda x: int(x * 2))

    lsa = LsiModel.load(os.path.join(BIGDATA_PATH, 'lsa_tweets_5589798_2003588x200.pkl'))
    tfidf = TfidfModel(id2word=lsa.id2word, dictionary=lsa.id2word)
    bows = np.array([lsa.id2word.doc2bow(txt.split()) for txt in tweets.text])
    # tfidfs = tfidf[bows]

    X = pd.DataFrame([pd.Series(dict(v)) for v in tqdm(lsa[tfidf[bows]], total=len(bows))], index=tweets.index)

    mask = ~X.isnull().any(axis=1)
    mask.index = tweets.index
    # >>> sum(~mask)
    # 99
    # >>> tweets.loc[mask.argmin()]
    # isbot                 0.17
    # strict                  13
    # user      b'CrisParanoid:'
    # text         b'#sad again'
    # Name: 571, dtype: object

    X = X[mask]
    y = tweets.isbot[mask]
    labels = labels[mask]

    test_size = 1.0 - training_size if training_size < 1 else float(len(X) - training_size) / len(X)
    Xindex, Xindex_test, yindex, yindex_test = train_test_split(X.index.values, y.index.values, test_size=test_size)
    X, Xtest, y, ytest = X.loc[Xindex], X.loc[Xindex_test], y.loc[yindex], y.loc[yindex_test]
    
    # labels_test = labels.loc[yindex_test]
    labels = labels.loc[yindex]

    tsne = TSNE(metric='precomputed', n_components=n_components, angle=angle, perplexity=perplexity)
    tsne = tsne.fit(positive_distances(X.values, metric=metric))

    return tsne, X, Xtest, y, ytest

# lda.save('lda')

# svr_lin = SVR(kernel='linear', C=1e3)
# y_lin = svr_lin.fit(X, y).predict(Xtest)
# print(mean_squared_error(y_lin, ytest))
# svr_rbf = SVR(kernel='rbf', C=1e3, gamma=0.1)
# y_rbf = svr_rbf.fit(X, y).predict(Xtest)
# lda.save('svr_rbf')

# svr_poly = SVR(kernel='poly', C=1e3, degree=2)
# y_poly = svr_poly.fit(X, y).predict(Xtest)



from __future__ import print_function, unicode_literals, division, absolute_import
from builtins import int, round, str
from future import standard_library
standard_library.install_aliases()
from builtins import object  # NOQA

from gensim.models import Word2Vec

class W2V(object):
    w2v = None

    def __init__(self, path='/data/Googl'):




from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa
from builtins import *  # noqa

from nlpia.data.loaders import get_data, read_csv, untar, no_tqdm, dropbox_basename, download, download_file, multifile_dataframe  # noqa

__all__ = [s.strip() for s in 'get_data, read_csv, untar, no_tqdm, dropbox_basename, download, download_file, multifile_dataframe'.split(',')]



from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa
from builtins import *  # noqa

import os
import re
# import shutil
import requests
from nlpia.constants import logging, DATA_PATH, BIGDATA_PATH

from tqdm import tqdm
from pugnlp.futil import path_status
import pandas as pd
import tarfile

"""Loaders and downloaders for data files and models required for the examples in NLP in Action

>>> from nlpia.data import download
>> download()  Will take hours and 8GB of storage
"""
np = pd.np
logger = logging.getLogger(__name__)

W2V_FILE = 'GoogleNews-vectors-negative300.bin.gz'
BIG_URLS = {
    'w2v': (
        'https://www.dropbox.com/s/965dir4dje0hfi4/GoogleNews-vectors-negative300.bin.gz?dl=1',
        1647046227,
        ),
    'slang': (
        'https://www.dropbox.com/s/43c22018fbfzypd/slang.csv.gz?dl=1',
        117633024,
        ),
    'tweets': (
        'https://www.dropbox.com/s/5gpb43c494mc8p0/tweets.csv.gz?dl=1',
        311725313,
        ),
    'lsa_tweets': (
        'https://www.dropbox.com/s/rpjt0d060t4n1mr/lsa_tweets_5589798_2003588x200.tar.gz?dl=1',
        3112841563,  # 3112841312,
        ),
    'imdb': (
        'https://www.dropbox.com/s/yviic64qv84x73j/aclImdb_v1.tar.gz?dl=1',
        3112841563,  # 3112841312,
        ),
    }
DATA_NAMES = {
    'pointcloud': os.path.join(DATA_PATH, 'pointcloud.csv.gz')
}

W2V_PATH = os.path.join(BIGDATA_PATH, W2V_FILE)
TEXTS = ['kite_text.txt', 'kite_history.txt']
CSVS = ['mavis-batey-greetings.csv', 'sms-spam.csv']


def untar(fname):
    if fname.endswith("tar.gz"):
        with tarfile.open(fname) as tf:
            tf.extractall()
    else:
        print("Not a tar.gz file: {}".format(fname))


for filename in TEXTS:
    with open(os.path.join(DATA_PATH, filename)) as f:
        locals()[filename.split('.')[0]] = f.read()


def read_csv(*args, **kwargs):
    """Like pandas.read_csv, only little smarter (checks first column to see if it should be the data frame index)

    >>> read_csv('mavis-batey-greetings.csv').head()
    """
    index_names = ('Unnamed: 0', 'pk', 'index', '')
    kwargs.update({'low_memory': False})
    df = pd.read_csv(*args, **kwargs)
    if ((df.columns[0] in index_names or (df[df.columns[0]] == df.index).all()) or
        (df[df.columns[0]] == np.arange(len(df))).all() or
        ((df.index == np.arange(len(df))).all() and str(df[df.columns[0]].dtype).startswith('int') and
         df[df.columns[0]].count() == len(df))):
        df = df.set_index(df.columns[0], drop=True)
        if df.index.name in ('Unnamed: 0', ''):
            df.index.name = None
    return df


for filename in CSVS:
    locals()['df_' + filename.split('.')[0].replace('-', '_')] = read_csv(os.path.join(DATA_PATH, filename))


harry_docs = ["The faster Harry got to the store, the faster and faster Harry would get home.",
              "Harry is hairy and faster than Jill.",
              "Jill is not as hairy as Harry."]


def no_tqdm(it, total=1):
    return it


def dropbox_basename(url):
    filename = os.path.basename(url)
    match = re.findall(r'\?dl=[0-9]$', filename)
    if match:
        return filename[:-len(match[0])]
    return filename


def download(names=None, verbose=True):
    names = [names] if isinstance(names, (str, bytes)) else names
    names = names or BIG_URLS.keys()
    file_paths = {}
    for name in names:
        name = name.lower().strip()
        if name in BIG_URLS:
            file_paths[name] = download_file(BIG_URLS[name][0],
                                             data_path=BIGDATA_PATH,
                                             size=BIG_URLS[name][1],
                                             verbose=verbose)
            if file_paths[name].endswith('.tar.gz'):
                untar(file_paths[name])
            file_paths[name] = file_paths[name][:-7]  # FIXME: rename tar.gz file so that it mimics contents
    return file_paths


def download_file(url, data_path=BIGDATA_PATH, filename=None, size=None, chunk_size=4096, verbose=True):
    """Uses stream=True and a reasonable chunk size to be able to download large (GB) files over https"""
    if filename is None:
        filename = dropbox_basesname(url)
    file_path = os.path.join(data_path, filename)
    if url.endswith('?dl=0'):
        url = url[:-1] + '1'  # noninteractive download
    if verbose:
        tqdm_prog = tqdm
        print('requesting URL: {}'.format(url))
    else:
        tqdm_prog = no_tqdm
    r = requests.get(url, stream=True, allow_redirects=True)
    size = r.headers.get('Content-Length', None) if size is None else size
    print('remote size: {}'.format(size))

    stat = path_status(file_path)
    print('local size: {}'.format(stat.get('size', None)))
    if stat['type'] == 'file' and stat['size'] == size:  # TODO: check md5 or get the right size of remote file
        r.close()
        return file_path

    print('Downloading to {}'.format(file_path))

    with open(file_path, 'wb') as f:
        for chunk in tqdm_prog(r.iter_content(chunk_size=chunk_size)):
            if chunk:  # filter out keep-alive chunks
                f.write(chunk)

    r.close()
    return file_path


def multifile_dataframe(paths=['urbanslang{}of4.csv'.format(i) for i in range(1, 5)], header=0, index_col=None):
    """Like pandas.read_csv, but loads and concatenates (df.append(df)s) DataFrames together"""
    df = pd.DataFrame()
    for p in paths:
        df = df.append(read_csv(p, header=header, index_col=index_col), ignore_index=True if not index_col else False)
    if index_col and df.index.name == index_col:
        del df[index_col]
    return df


def get_data(name='sms-spam'):
    try:
        return read_csv(os.path.join(DATA_PATH, name + '.csv.gz'))
    except IOError:
        try:
            return read_csv(os.path.join(DATA_PATH, name + '.csv'))
        except IOError:
            logger.error('Unable to find dataset named {} in DATA_PATH'.format(name))
            raise



# from plotly.offline import download_plotlyjs, init_notebook_mode, plot, iplot
# import plotly.graph_objs as go
from past.builtins import basestring

from seaborn import plt
from mpl_toolkits.mplot3d import Axes3D  # noqa

import plotly.plotly as plotly

import pandas as pd

np = pd.np


def scatter_3d(df, labels=None, depthshade=True):
    df = getattr(df, 'embedding_', df)
    labels = df[labels] if (isinstance(labels, *(int, basestring)) and
                            labels in getattr(df, 'columns', set())) else labels
    labels = np.array(np.zeros(shape=(len(df),)) if labels is None else labels)
    if isinstance(labels[0], float):
        labels = labels.astype(int)  # TODO: LabelEncoder
    if str(labels.dtype).startswith('int'):
        labels = np.array(list('grbkcym'))[labels % 7]

    try:
        df = df.values
    except:
        pass

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(df[:, 0], df[:, 1], df[:, 2], zdir='z', s=20, c=labels, depthshade=depthshade)
    return fig


def offline_plotly(df=None):
    df = df or pd.read_csv('https://raw.githubusercontent.com/plotly/datasets/master/iris.csv')
    print(df.head())

    data = []
    clusters = []
    colors = ['rgb(228,26,28)', 'rgb(55,126,184)', 'rgb(77,175,74)']

    for i in range(len(df['Name'].unique())):
        name = df['Name'].unique()[i]
        color = colors[i]
        x = df[df['Name'] == name ]['SepalLength']
        y = df[df['Name'] == name ]['SepalWidth']
        z = df[df['Name'] == name ]['PetalLength']

        trace = dict(
            name = name,
            x = x, y = y, z = z,
            type = "scatter3d",    
            mode = 'markers',
            marker = dict( size=3, color=color, line=dict(width=0) ) )
        data.append( trace )

    layout = dict(
        width=800,
        height=550,
        autosize=False,
        title='Iris dataset',
        scene=dict(
            xaxis=dict(
                gridcolor='rgb(255, 255, 255)',
                zerolinecolor='rgb(255, 255, 255)',
                showbackground=True,
                backgroundcolor='rgb(230, 230,230)'
            ),
            yaxis=dict(
                gridcolor='rgb(255, 255, 255)',
                zerolinecolor='rgb(255, 255, 255)',
                showbackground=True,
                backgroundcolor='rgb(230, 230,230)'
            ),
            zaxis=dict(
                gridcolor='rgb(255, 255, 255)',
                zerolinecolor='rgb(255, 255, 255)',
                showbackground=True,
                backgroundcolor='rgb(230, 230,230)'
            ),
            aspectratio = dict( x=1, y=1, z=0.7 ),
            aspectmode = 'manual'        
        ),
    )

    fig = dict(data=data, layout=layout)

    # IPython notebook
    # plotly.iplot(fig, filename='pandas-3d-iris', validate=False)

    url = plotly.offline.plot(fig, filename='pandas-3d-iris', validate=False)
